import pandas as pd
from pydantic import BaseModel, Field


class Aggregation(BaseModel):
    open: str = Field(default="first")
    """"""
    high: str = Field(default="max")
    """"""
    low: str = Field(default="min")
    """"""
    close: str = Field(default="last")
    """"""
    volume: str = Field(default="sum")
    """"""


def resample(
    df: pd.DataFrame, tf: str = "1H", agg: Aggregation = Aggregation(), on: str = "date"
) -> pd.DataFrame:
    """
    Resamples a DataFrame over a specified time frequency and applies aggregation functions.

    This function resamples the input DataFrame `df` using the time frequency provided by `tf`.
    It then applies the specified aggregation functions on the relevant columns and fills any
    missing values via forward fill.

    Parameters:
        df: The DataFrame to be resampled. It must contain a time-based column that can be used
            for resampling.
        tf: The time frequency for resampling, default is "1H" (one hour). It should be in a format
            accepted by pandas' `resample()` function, such as "1D" for daily, "1T" for minute, etc.
        agg: A dictionary defining the aggregation method for each relevant column. The default value
            aggregates the "open" column by the first value, "high" by the max, "low" by the min,
            "close" by the last value, and "volume" by the sum.
        on: The name of the column to be used for resampling, typically a date or timestamp column.
            The default is "date".

    Returns:
        A resampled DataFrame with the aggregation applied, missing values dropped, and any remaining missing values forward-filled.

    Examples:

    ```py
    import pandas as pd

    df = pd.DataFrame({
         "date": pd.date_range(start="2023-01-01", periods=5, freq="1H"),
         "open": [100, 101, 102, 103, 104],
         "high": [110, 111, 112, 113, 114],
         "low": [90, 91, 92, 93, 94],
         "close": [105, 106, 107, 108, 109],
         "volume": [1000, 1500, 1200, 1100, 1400]
     })
     resample(df, tf="2H", on="date")
    ```
    """
    return df.resample(tf, on=on).agg(agg).dropna(how="all").fillna(method="ffill")  # type: ignore



import pandas as pd
import clickhouse_connect
from pydantic import BaseModel, ConfigDict
from clickhouse_connect.driver.client import Client
from .generate_table_schema import generate_table_schema


class BaseCH(BaseModel):
    host: str = "localhost"
    port: int = 8123
    database: str = "default"
    user: str = "default"
    password: str = ""


class ClickhouseSync(BaseModel):
    """
    An sync clickhouse client.

    Usage:
        ```python
        from tradingtoolbox.clickhouse import ClickhouseSync

        client = ClickhouseSync.create(
            host="localhost",
            port=8123,
            user="default",
            password="",
            database="default",
        )

        # Or you can create it without passing anything, and it will use the above defaults
        async_client = ClickhouseSync.create()
        ```
    """

    client: Client

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_assignment=True,
    )

    @classmethod
    def create(cls, **kwargs):
        base = BaseCH(**kwargs)
        kwargs["client"] = clickhouse_connect.get_client(
            host=base.host,
            port=base.port,
            username=base.user,
            password=base.password,
            database=base.database,
        )
        self = cls(**kwargs)

        return self

    def insert_df(self, df: pd.DataFrame, table_name: str, drop=False):
        """
        Parameters:
            df: The dataframe to insert
            table_name: The name of the table to insert into
            drop: Whether to drop the table if it already exists
        """
        if drop:
            self.drop_table(table_name)
        schema = generate_table_schema(df, table_name)

        self.client.command(schema)
        self.client.insert_df(table_name, df)

    def execute_command(self, msg: str):
        """
        Allows you to run any command you want

        Parameters:
            query: The query to execute
        """
        self.client.command(msg)

    # =============================================================== #

    def drop_table(self, table_name: str):
        """
        Drops a table from the database

        Parameters:
            table_name: The table's name
        """
        return self.execute_command(f"DROP TABLE IF EXISTS {table_name}")

    def create_table(
        self,
        table_name: str,
        schema: str,
        partition_key: str = "",
        order_by: str = "",
        primary_key: str = "",
        drop: bool = False,
    ):
        """
        Create a table

        Parameters:
            table_name: The name of the table
            schema: The schema of the table
            partition_key: The partition key
            order_by: The order by
            primary_key: The primary key
            drop: Whether to drop the table if it already exists
        """
        if drop:
            self.drop_table(table_name)

        query_string = f"""
CREATE TABLE IF NOT EXISTS {table_name} (
    {schema}
)
ENGINE = ReplacingMergeTree()
{partition_key}
{order_by}
{primary_key}
"""
        # print(query_string)
        return self.execute_command(query_string)

    def optimize_table(self, table_name: str):
        """
        Optimizes a table

        Parameters:
            table_name: The name of the table
        """
        command = f"OPTIMIZE TABLE {table_name} FINAL"
        return self.execute_command(command)

    def query(self, query: str):
        """
        Runs a query

        Parameters:
            query: The query to run
        """
        return self.client.query(query)

    def query_df(self, query: str):
        """
        Runs a query and returns the results as a dataframe

        Parameters:
            query: The query to run
        """
        return self.client.query_df(query)



from tradingtoolbox import print


def test_main():
    # os.remove("./logs/logs.log")
    print("Hello")

    # logger.warning("Warning")
    # logger.error("Error message")

    a = 4
    assert a == 4



from binance.client import Client
from enum import Enum
import pandas as pd
import os

from pydantic import BaseModel, ConfigDict


class Timeframes(Enum):
    """
    Timeframes for Binance

    Usage:

    Attributes:
        TF_1MINUTE (Literal): Binance 1 minute interval.
        TF_3MINUTE (Literal): Binance 3 minute interval.
        TF_5MINUTE (Literal): Binance 5 minute interval.
        TF_15MINUTE (Literal): Binance 15 minute interval.
        TF_30MINUTE (Literal): Binance 30 minute interval.
        TF_1HOUR (Literal): Binance 1 hour interval.
        TF_2HOUR (Literal): Binance 2 hour interval.
        TF_4HOUR (Literal): Binance 4 hour interval.
        TF_6HOUR (Literal): Binance 6 hour interval.
        TF_8HOUR (Literal): Binance 8 hour interval.
        TF_12HOUR (Literal): Binance 12 hour interval.
        TF_1DAY (Literal): Binance 1 day interval.
        TF_3DAY (Literal): Binance 3 day interval.
        TF_1WEEK (Literal): Binance 1 week interval.
        TF_1MONTH (Literal): Binance 1 month interval.
    """

    TF_1MINUTE = Client.KLINE_INTERVAL_1MINUTE
    TF_3MINUTE = Client.KLINE_INTERVAL_3MINUTE
    TF_5MINUTE = Client.KLINE_INTERVAL_5MINUTE
    TF_15MINUTE = Client.KLINE_INTERVAL_15MINUTE
    TF_30MINUTE = Client.KLINE_INTERVAL_30MINUTE
    TF_1HOUR = Client.KLINE_INTERVAL_1HOUR
    TF_2HOUR = Client.KLINE_INTERVAL_2HOUR
    TF_4HOUR = Client.KLINE_INTERVAL_4HOUR
    TF_6HOUR = Client.KLINE_INTERVAL_6HOUR
    TF_8HOUR = Client.KLINE_INTERVAL_8HOUR
    TF_12HOUR = Client.KLINE_INTERVAL_12HOUR
    TF_1DAY = Client.KLINE_INTERVAL_1DAY
    TF_3DAY = Client.KLINE_INTERVAL_3DAY
    TF_1WEEK = Client.KLINE_INTERVAL_1WEEK
    TF_1MONTH = Client.KLINE_INTERVAL_1MONTH

    def __str__(self):
        return self.value


class BinanceKlines(BaseModel):
    """
    A class that allows downloading klines from Binance.

    Usage:
        ```python
        from tradingtoolbox.exchanges import BinanceKlines, Timeframes

        klines = BinanceKlines()
        df = klines.get_futures_klines(
            Timeframes.TF_1HOUR, asset="BTCUSDT", ago="1 day ago UTC"
        )
        ```
    """

    client: Client = Client()

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_assignment=True,
    )

    def _create_binance_dataframe(self, klines: list, set_index: bool):
        columns = [
            "Open Time",
            "Open",
            "High",
            "Low",
            "Close",
            "Volume",
            "Close time",
            "Quote asset volume",
            "Number of trades",
            "Taker buy base asset volume",
            "Taker buy quote asset volume",
            "Ignore",
        ]
        df = pd.DataFrame(
            klines,
            dtype=float,
            columns=columns,  # type: ignore
        )

        df["Open Time"] = pd.to_datetime(df["Open Time"], unit="ms")
        df.drop(
            columns=[
                "Close time",
                "Quote asset volume",
                "Number of trades",
                "Taker buy base asset volume",
                "Taker buy quote asset volume",
                "Ignore",
            ],
            inplace=True,
        )
        # Rename columns using a dictionary
        new_column_names = {
            "Open Time": "Date",
            "Open": "open",
            "High": "high",
            "Low": "low",
            "Close": "close",
            "Volume": "volume",
        }
        df = df.rename(columns=new_column_names)
        if set_index:
            df.set_index("Date", inplace=True)

        return df

    def _create_data_dir(self, directory="./data"):
        os.makedirs(directory, exist_ok=True)

    def _get_klines(
        self,
        asset_type: str,
        tf: Timeframes,
        asset="BTCUSDT",
        ago="1 day ago UTC",
        cache=True,
        data_directory="./data",
        set_index=True,
    ) -> pd.DataFrame:
        file = f"./data/binance_{asset_type}_{asset}_{tf}_klines.parquet"
        self._create_data_dir(directory=data_directory)
        if cache:
            self._create_data_dir(directory=data_directory)
            if os.path.exists(file):
                df = pd.read_parquet(file)
                return df
        if asset_type == "futures":
            klines = self.client.futures_historical_klines(asset, tf.value, ago)
        else:
            klines = self.client.get_historical_klines(asset, tf.value, ago)

        df = self._create_binance_dataframe(klines, set_index)
        df["asset"] = asset
        df["timeframe"] = tf.value
        df.to_parquet(file)
        return df

    def get_futures_klines(
        self,
        tf: Timeframes,
        asset: str = "BTCUSDT",
        ago: str = "1 day ago UTC",
        cache: bool = True,
        data_directory: str = "./data",
        set_index: bool = True,
    ) -> pd.DataFrame:
        """
        Fetches the Futures (Perp Swaps) klines

        Parameters:
            tf: The timeframe to fetch.
            asset: The asset name.
            ago: The time ago you want to fetch it. Ex: '30 day ago UTC', or '3 months ago UTC'
            cache: Whether to cache the data. Subsequent calls will **not** call the API.
            data_directory: The directory where the klines fill be stored as parquet files
            set_index: Whether to set the index of the returned DataFrame. If True, the 'Date' column will be set as index
        """
        return self._get_klines(
            asset_type="futures",
            tf=tf,
            asset=asset,
            ago=ago,
            cache=cache,
            data_directory=data_directory,
            set_index=set_index,
        )

    def get_klines(
        self,
        tf: Timeframes,
        asset="BTCUSDT",
        ago="1 day ago UTC",
        cache=True,
        data_directory="./data",
        set_index=True,
    ) -> pd.DataFrame:
        return self._get_klines(
            asset_type="spot",
            tf=tf,
            asset=asset,
            ago=ago,
            cache=cache,
            data_directory=data_directory,
            set_index=set_index,
        )
